@@ -686,6 +686,10 @@ divll_pre(ulong a_lo, ulong n, ulong ninv)
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                66
 #define __F2x_MUL_KARATSUBA_LIMIT        15
+#define __F2xqX_BARRETT_LIMIT            48
+#define __F2xqX_DIVREM_BARRETT_LIMIT     292
+#define __F2xqX_INVBARRETT_LIMIT         46
+#define __F2xqX_REM_BARRETT_LIMIT        292
 #define __Flx_BARRETT_HALFMULII_LIMIT    21
 #define __Flx_BARRETT_KARATSUBA_LIMIT    1172
 #define __Flx_BARRETT_MULII2_LIMIT       16
@@ -744,16 +748,21 @@ divll_pre(ulong a_lo, ulong n, ulong ninv)
 #define __LOGAGM_LIMIT                   6
 #define __MULII_FFT_LIMIT                -1
 #define __MULII_KARATSUBA_LIMIT          -1
-#define __MULRR_MULII_LIMIT              74
+#define __MULRR_MULII_LIMIT              55
 #define __RgX_MUL_LIMIT                  9
 #define __RgX_SQR_LIMIT                  38
 #define __SQRI_FFT_LIMIT                 -1
 #define __SQRI_KARATSUBA_LIMIT           -1
+#define __SQRR_SQRI_LIMIT                12
 #else
 #define __AGM_ATAN_LIMIT                 89
 #define __DIVRR_GMP_LIMIT                4
 #define __EXPNEWTON_LIMIT                197
 #define __F2x_MUL_KARATSUBA_LIMIT        23
+#define __F2xqX_BARRETT_LIMIT            17
+#define __F2xqX_DIVREM_BARRETT_LIMIT     46
+#define __F2xqX_INVBARRETT_LIMIT         22
+#define __F2xqX_REM_BARRETT_LIMIT        48
 #define __Flx_BARRETT_HALFMULII_LIMIT    23
 #define __Flx_BARRETT_KARATSUBA_LIMIT    905
 #define __Flx_BARRETT_MULII2_LIMIT       647
@@ -812,11 +821,12 @@ divll_pre(ulong a_lo, ulong n, ulong ninv)
 #define __LOGAGM_LIMIT                   45
 #define __MULII_FFT_LIMIT                -1
 #define __MULII_KARATSUBA_LIMIT          -1
-#define __MULRR_MULII_LIMIT              8
+#define __MULRR_MULII_LIMIT              19
 #define __RgX_MUL_LIMIT                  7
 #define __RgX_SQR_LIMIT                  34
 #define __SQRI_FFT_LIMIT                 -1
 #define __SQRI_KARATSUBA_LIMIT           -1
+#define __SQRR_SQRI_LIMIT                9
 #endif
 #line 2 "../src/kernel/gmp/int.h"
 /* Copyright (C) 2000  The PARI group.
@@ -964,7 +974,7 @@ new_chunk(size_t x) /* x is a number of longs */
     new_chunk_resize(x);
   avma = (pari_sp)z;
 #ifdef MEMSTEP
-  if (DEBUGMEM && pari_mainstack->memused != DISABLE_MEMUSED) {
+  if (DEBUGMEM>1 && pari_mainstack->memused != DISABLE_MEMUSED) {
     long d = (long)pari_mainstack->memused - (long)z;
     if (labs(d) > 4*MEMSTEP)
     {
@@ -984,6 +994,14 @@ stack_malloc(size_t N)
   return (char*)new_chunk(n);
 }
 
+INLINE char *
+stack_malloc_align(size_t N, long k)
+{
+  ulong d = ((ulong)avma) % k;
+  if (d) (void)new_chunk(d/sizeof(long));
+  return (char*) new_chunk(nchar2nlong(N));
+}
+
 INLINE char *
 stack_calloc(size_t N)
 {
@@ -1096,6 +1114,8 @@ INLINE GEN
 mpneg(GEN x) { GEN y = leafcopy(x); togglesign(y); return y; }
 INLINE GEN negi(GEN x) { return mpneg(x); }
 INLINE GEN negr(GEN x) { return mpneg(x); }
+INLINE GEN gmax_shallow(GEN x, GEN y) { return gcmp(x,y)<0? y: x; }
+INLINE GEN gmin_shallow(GEN x, GEN y) { return gcmp(x,y)<0? x: y; }
 
 /* negate in place */
 INLINE void
@@ -1195,7 +1215,7 @@ itou(GEN x)
     case 3: return x[2];
     default:
       pari_err_OVERFLOW("t_INT-->ulong assignment");
-      return 0; /* not reached */
+      return 0; /* LCOV_EXCL_LINE */
   }
 }
 
@@ -1206,6 +1226,24 @@ itou_or_0(GEN x) {
   return (ulong)x[2];
 }
 
+INLINE ulong
+umuluu_or_0(ulong x, ulong y)
+{
+  ulong z;
+  LOCAL_HIREMAINDER;
+  z = mulll(x, y);
+  return hiremainder? 0: z;
+}
+/* return x*y if <= n, else 0. Beware overflow */
+INLINE ulong
+umuluu_le(ulong x, ulong y, ulong n)
+{
+  ulong z;
+  LOCAL_HIREMAINDER;
+  z = mulll(x, y);
+  return (hiremainder || z > n)? 0: z;
+}
+
 INLINE GEN
 real_0_bit(long bitprec) { GEN x=cgetr(2); x[1]=evalexpo(bitprec); return x; }
 INLINE GEN
@@ -1283,6 +1321,19 @@ cmpsr(long x, GEN y)
 }
 INLINE int
 cmprs(GEN x, long y) { return -cmpsr(y,x); }
+/* compare x and y */
+INLINE int
+cmpui(ulong x, GEN y)
+{
+  ulong p;
+  if (!x) return -signe(y);
+  if (signe(y) <= 0) return 1;
+  if (lgefint(y) > 3) return -1;
+  p = y[2]; if (p == x) return 0;
+  return p < x ? 1 : -1;
+}
+INLINE int
+cmpiu(GEN x, ulong y) { return -cmpui(y,x); }
 /* compare x and |y| */
 INLINE int
 abscmpui(ulong x, GEN y)
@@ -1328,6 +1379,14 @@ mpcmp(GEN x, GEN y)
   return (typ(y)==t_INT) ? -cmpir(y,x) : cmprr(x,y);
 }
 
+/* x == y ? */
+INLINE int
+equalui(ulong x, GEN y)
+{
+  if (!x) return !signe(y);
+  if (signe(y) <= 0 || lgefint(y) != 3) return 0;
+  return ((ulong)y[2] == (ulong)x);
+}
 /* x == y ? */
 INLINE int
 equalsi(long x, GEN y)
@@ -1352,6 +1411,8 @@ INLINE int
 absequaliu(GEN x, ulong y) { return absequalui(y,x); }
 INLINE int
 equalis(GEN x, long y) { return equalsi(y,x); }
+INLINE int
+equaliu(GEN x, ulong y) { return equalui(y,x); }
 
 /* assume x != 0, is |x| == 2^n ? */
 INLINE int
@@ -1528,14 +1589,13 @@ ceildivuu(ulong a, ulong b)
   return (a%b)? c+1: c;
 }
 
-
 INLINE ulong
-udivui_rem(ulong x, GEN y, ulong *r)
+uabsdivui_rem(ulong x, GEN y, ulong *r)
 {
   long q, s = signe(y);
   LOCAL_HIREMAINDER;
 
-  if (!s) pari_err_INV("udivui_rem",gen_0);
+  if (!s) pari_err_INV("uabsdivui_rem",gen_0);
   if (!x || lgefint(y)>3) { *r = x; return 0; }
   hiremainder=0; q = (long)divll(x, (ulong)y[2]);
   if (s < 0) q = -q;
@@ -1545,7 +1605,7 @@ udivui_rem(ulong x, GEN y, ulong *r)
 /* assume d != 0 and |n| / d can be represented as an ulong.
  * Return |n|/d, set *r = |n| % d */
 INLINE ulong
-udiviu_rem(GEN n, ulong d, ulong *r)
+uabsdiviu_rem(GEN n, ulong d, ulong *r)
 {
   switch(lgefint(n))
   {
@@ -1647,6 +1707,11 @@ umodui(ulong x, GEN y)
   return x % (ulong)y[2];
 }
 
+INLINE ulong
+ugcdiu(GEN x, ulong y) { return ugcd(umodiu(x,y), y); }
+INLINE ulong
+ugcdui(ulong y, GEN x) { return ugcd(umodiu(x,y), y); }
+
 INLINE GEN
 remsi(long x, GEN y)
 { long r; (void)sdivsi_rem(x,y, &r); return stoi(r); }
@@ -1700,9 +1765,8 @@ rdiviiz(GEN x, GEN y, GEN z)
 INLINE GEN
 rdivii(GEN x, GEN y, long prec)
 {
-  GEN z = cgetr(prec);
+  GEN z = itor(x, prec);
   pari_sp av = avma;
-  affir(x, z);
   if (lg(y) == 3) {
     affrr(divru(z, y[2]), z);
     if (signe(y) < 0) togglesign(z);
@@ -1756,7 +1820,7 @@ dvdiuz(GEN x, ulong y, GEN z)
 {
   const pari_sp av = avma;
   ulong r;
-  GEN p1 = diviu_rem(x,y, &r);
+  GEN p1 = absdiviu_rem(x,y, &r);
   avma = av; if (r) return 0;
   affii(p1,z); return 1;
 }
@@ -1797,7 +1861,7 @@ Fl_mul_pre(ulong a, ulong b, ulong p, ulong pi)
 }
 
 INLINE ulong
-Fl_addmul_pre(ulong x0, ulong x1, ulong y0, ulong p, ulong pi)
+Fl_addmul_pre(ulong y0, ulong x0, ulong x1, ulong p, ulong pi)
 {
   ulong l0, h0;
   LOCAL_HIREMAINDER;
@@ -2181,11 +2245,11 @@ INLINE void divssz(long x, long y, GEN z)
 { affsi(x/y, z); }
 
 INLINE void modisz(GEN y, long s, GEN z)
-{ pari_sp av = avma; affii(modis(y,s),z); avma = av; }
+{ affsi(smodis(y,s),z); }
 INLINE void modsiz(long s, GEN y, GEN z)
 { pari_sp av = avma; affii(modsi(s,y),z); avma = av; }
 INLINE void modssz(long s, long y, GEN z)
-{ pari_sp av = avma; affii(modss(s,y),z); avma = av; }
+{ affsi(smodss(s,y),z); }
 
 INLINE void mpaddz(GEN x, GEN y, GEN z)
 { pari_sp av = avma; mpaff(mpadd(x,y),z); avma = av; }
